/*

DESCRIPTION:    PIC-specific constants (PIC peripheral, analog channel, ADC,
                NVM, PPS, and pin mapping), object definitions, and function
                declarations.

COPYRIGHT:      (c) 2022, 2023, Brian Cornell, all rights reserved.

 TARGET MCU:    PIC16F1776

MODIFICATION HISTORY:

    12 Jan 23, BWC, added SHARED ADC & VMV CONFIGURATION & object defintions,
    function declarations.
    
    9 Mar 22, BWC, created.


*/


#define _PIC16F1776_MAP


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * PERIPHERAL ALIASES.
 * Generic constants that alias peripherals to function.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define S_AL_PWM_LD         PWMLD           //PWM load register
#define S_AL_REG_STATUS     STATUS          //Status register
#define S_AL_REG_PCON       PCON            //Status register
#define S_AL_WDT_ENABLE     SWDTEN          //Watchdog enable bit


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * ANALOG CHANNEL MAPPING. channel=port=pin. The values assigned to these
 * constants should match those used by ADC.
 * 
 * IMPORTANT: these values are 0-offset meaning they are to be assigned to ADC
 * channel using ADCON0bits.CHS.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define ANLCAN0             0x00            //AN0=RA0=2
#define ANLCAN1             0x01            //AN1=RA1=3
#define ANLCAN2             0x02            //AN2=RA2=4
#define ANLCAN3             0x03            //AN3=RA3=5
#define ANLCAN4             0x04            //AN4=RA5=7
#define ANLCAN8             0x08            //AN8=RB2=23
#define ANLCAN9             0x09            //AN9=RB3=24
#define ANLCAN10            0x0A            //AN10=RB1=22
#define ANLCAN11            0x0B            //AN11=RB4=25
#define ANLCAN12            0x0C            //AN12=RB0=21
#define ANLCAN13            0x0D            //AN13=RB5=26
#define ANLCAN14            0x0E            //AN14=RC2=13
#define ANLCAN15            0x0F            //AN15=RC3=14
#define ANLCAN16            0x10            //AN16=RC4=15
#define ANLCAN17            0x11            //AN17=RC5=16
#define ANLCAN18            0x12            //AN18=RC6=17
#define ANLCAN19            0x13            //AN19=RC7=18
#define ANLCANS1            0x21            //Switched AN1=RA1=3
#define ANLCANS10           0x2A            //Switched AN10=RB1=22
#define ANLCANS18           0x32            //Switched AN18=RC6=17

//Internal channels (peripherals)
#define ADC_CH_FVR          0xFC            //FVR
#define ADC_CH_DAC_1        0xF8            //DAC1
#define ADC_CH_TMP          0xF4            //Temperature indicator
#define ADC_CH_DAC_2        0xF0            //DAC2
#define ADC_CH_DAC_3        0xEC            //DAC3
#define ADC_CH_DAC_4        0xE8            //DAC4
#define ADC_CH_DAC_5        0xE4            //DAC5
#define ADC_CH_DAC_7        0xDC            //DAC7


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * PPS INPUT PIN DEFINITIONS. Use to assign peripheral to input pin.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define PPSRA0              0x00            //RA0
#define PPSRA1              0x01            //RA1
#define PPSRA2              0x02            //RA2
#define PPSRA3              0x03            //RA3
#define PPSRA4              0x04            //RA4
#define PPSRA5              0x05            //RA5
#define PPSRA6              0x06            //RA6
#define PPSRA7              0x07            //RA7
#define PPSRB0              0x08            //RB0
#define PPSRB1              0x09            //RB1
#define PPSRB2              0x0A            //RB2
#define PPSRB3              0x0B            //RB3
#define PPSRB4              0x0C            //RB4
#define PPSRB5              0x0D            //RB5
#define PPSRB6              0x0E            //RB6
#define PPSRB7              0x0F            //RB7
#define PPSRC0              0x10            //RC0
#define PPSRC1              0x11            //RC1
#define PPSRC2              0x12            //RC2
#define PPSRC3              0x13            //RC3
#define PPSRC4              0x14            //RC4
#define PPSRC5              0x15            //RC5
#define PPSRC6              0x16            //RC6
#define PPSRC7              0x17            //RC7

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * PPS OUTPUT PERHIPERAL MAPPING. Use to assign output pin to peripheral.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define PPS_MD3_OUT         0x30            //Data signal modulator 3
#define PPS_MD2_OUT         0x2F            //Data signal modulator 2
#define PPS_MD1_OUT         0x2E            //Data signal modulator 1
#define PPS_C6_OUT          0x2B            //Comparator 6
#define PPS_C5_OUT          0x2A            //Comparator 5
#define PPS_C4_OUT          0x29            //Comparator 4
#define PPS_C3_OUT          0x28            //Comparator 3
#define PPS_C2_OUT          0x27            //Comparator 2
#define PPS_C1_OUT          0x26            //Comparator 1
#define PPS_DT_OUT          0x25            //EUSART sync. data transmit
#define PPS_TX_OUT          0x24            //EUSART async. transmit
#define PPS_SDO_OUT         0x23            //SPI serial data out
#define PPS_SDA_OUT         0x22            //I2C serial data
#define PPS_SCL_OUT         0x21            //I2C clock
#define PPS_PWM11_OUT       0x1F            //PWM 11
#define PPS_PWM6_OUT        0x1E            //PWM 6
#define PPS_PWM5_OUT        0x1D            //PWM 5
#define PPS_PWM9_OUT        0x1B            //PWM 9
#define PPS_PWM4_OUT        0x1A            //PWM 4
#define PPS_PWM3_OUT        0x19            //PWM 3
#define PPS_CCP7_OUT        0x17            //Capture compare PWM 7
#define PPS_CCP2_OUT        0x16            //Capture compare PWM 2
#define PPS_CCP1_OUT        0x15            //Capture compare PWM 1
#define PPS_COG3D_OUT       0x10            //Complementary output gen. 3
#define PPS_COG3C_OUT       0x0F            //Complementary output gen. 3
#define PPS_COG3B_OUT       0x0E            //Complementary output gen. 3
#define PPS_COG3A_OUT       0x0D            //Complementary output gen. 3
#define PPS_COG2D_OUT       0x0C            //Complementary output gen. 2
#define PPS_COG2C_OUT       0x0B            //Complementary output gen. 2
#define PPS_COG2B_OUT       0x0A            //Complementary output gen. 2
#define PPS_COG2A_OUT       0x09            //Complementary output gen. 2
#define PPS_COG1D_OUT       0x08            //Complementary output gen. 1
#define PPS_COG1C_OUT       0x07            //Complementary output gen. 1
#define PPS_COG1B_OUT       0x06            //Complementary output gen. 1
#define PPS_COG1A_OUT       0x05            //Complementary output gen. 1
#define PPS_LC4_OUT         0x04            //Logic cell 4
#define PPS_LC3_OUT         0x03            //Logic cell 3
#define PPS_LC2_OUT         0x02            //Logic cell 2
#define PPS_LC1_OUT         0x01            //Logic cell 1
#define PPS_LAT_OUT         0x00            //Port / pin latch (no peripheral)


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * FLASH NVM DEFINITIONS.
 * 
 * Important:  flash architecture is PIC family dependent and the memory map /
 * constants must be reviewed when changing MCUs.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define S_SZ_FLASHR         32              		      //Row size
#define S_SZ_FLASHC         (S_SZ_FLASHR-sizeof(uint16_t))      //Size w/o cksum
#define S_SZ_FLASHW         (S_SZ_FLASHR / sizeof(uint16_t))    //words size
#define S_SZ_FLASHS         (S_SZ_FLASHC / sizeof(uint24_t))    //swords size

//Checksum stored in little endian format in last word of flash row
#define S_AR_FLCS_L         S_SZ_FLASHC     //Low byte address of checksum word
#define S_AR_FLCS_H         (S_SZ_FLASHC+1) //High byte address of checksum word
#define S_AR_FLCS_W         (S_SZ_FLASHW-1) //16-bit address of checksum word

//Flash memory unlock sequence macro
#define unlockSeq() PMCON2=0x55; PMCON2=0xAA; WR=TRUE; NOP(); NOP()


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * FVR CONFIGURATION CONSTANTS.
 * 
 * Used with ADC scheme documented below and function ReadVoltage.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define CFG_FVR_ADC_4X      0x03            //ADC buffer set to 4x gain
#define CFG_FVR_ADC_2X      0x02            //ADC buffer set to 2x gain
#define FI_FVR_DLY          30              //uS for FVR buffer to stabilize


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * SHARED ADC & VMV CONFIGURATION.
 * 
 * Implementation specific constants go in project includes.
 * 
 * ADC configuration, acquisition, and value storage: this section defines a 
 * uniform method for data acquisition and storage. The ADC is configured to 
 * operate thru function ReadVoltage and should not be accessed directly once 
 * it has been Initialized [at boot]. Acquired values are stored in an 
 * object of type volts. The integer and decimal portions of a voltage are 
 * stored in discrete variables.
 * 
 * Since the PIC family lacks floating point capability a 16-bit unsigned int
 * variable can represent the combined integer & decimal components in a format
 * referred to as 'VMV' (Volts-MilliVolts) format. This allows evaluation as
 * an integer with the benefit of including the decimal portion for improved 
 * resolution. This is done by multiplying the integer and dividing the decimal
 * values by some factors of 10, in effect shifting integer values left and
 * truncating decimal positions.
 * 
 * More places shifted offers higher resolution but with more limited range
 * (e.g. overflow). The constants S_RT_VMV_INT & S_RT_VMV_DEC define the use of 
 * VMV values and must be assigned the paired values:
 * # Decimal Places         S_RT_VMV_INT        S_RT_VMV_DEC
 *          0               1                   1000
 *          1               10                  100
 *          2               100                 10
 *          3               1000                1
 * 
 * Prior to calling function ReadVoltage the variable mvr of the volts object
 * passed to the function must be initialized to S_RT_ADC_MV_M, S_RT_ADC_MV_H,
 *  or S_RT_ADC_MV_L. This defines the Vref (5.0, 4.096, or 2.048) used and 
 * trades resolution for measurement voltage range.
 * 
 * When S_RT_ADC_MV_M is used the ADC must use a 5V positive reference. The
 * constant S_MK_ADC_5V must be set to either S_MK_ADC_VDD or S_MK_ADC_VREF. The
 * 5V source should be stable and low-noise: a full scale error of 120mV (or ~ 
 * 117uV/bit) is introduced because 5V is not evenly divisible by the ADC's bit 
 * resolution. This is in addition to the ADC's accuracy specification.
 * 
 * The constant S_RT_MV_M_COR is used by the function ReadVoltage to
 * compensate for this but does not fully eliminate the error. The value should 
 * be adjusted to obtain the most accurate values for the scale of the voltages 
 * being measured. The starting value is derived by dividing the integer 
 * resolution (5mV) by the 1 LSB error (0.005 - 0.00488281 = 0.000117).
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define S_EN_ADC            0x01            //Enable bit in ADCON0 register
#define S_MK_ADC_VDD        0b00            //Mask for +Vref set to Vdd pin
#define S_MK_ADC_VREF       0b10            //Mask for +Vref set to Vref pin
#define S_MK_ADC_FVR        0b11            //Set Vref=FVR
#define S_LM_ADC_VAL        1023            //Maximum 10-bit value
#define S_QY_MV_V           1000            //Number of mV in one volt
#define S_RT_ADC_MV_M       5               //mV / ADC bit for maximum range
#define S_RT_ADC_MV_H       4               //mV / ADC bit for high range
#define S_RT_ADC_MV_L       2               //mV / ADC bit for low range
#define S_RT_ADC_MV_U       0               //Undefined range
#define S_QY_ADC_1V_M       (S_QY_MV_V / S_RT_ADC_MV_M) //1V for max
#define S_QY_ADC_1V_H       (S_QY_MV_V / S_RT_ADC_MV_H) //1V for high
#define S_QY_ADC_1V_L       (S_QY_MV_V / S_RT_ADC_MV_L) //1V low
#define S_SZ_ADC_INT        sizeof(uint16_t)    //# bytes in ADC integer
#define POS_ADC_HB          1               //ADC result high byte
#define POS_ADC_LB          0               //ADC result low byte


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * OBJECT DEFINITIONS.
 *  +   adcresult=Object to retrieve ADC result.
 *  +   nvmcfg=NVM configuration row object.
 *  +	volts=used by function ReadVoltage to store an ADC measured voltage with
 *      up to three decimal places resolution.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
typedef union {
    unsigned char   result[S_SZ_ADC_INT];   //Low, high byte in little endian
    uint16_t        value;                  //16-bit integer value
} adcresult;

typedef union {
    uint8_t         byte[S_SZ_FLASHR];
    uint16_t        word[S_SZ_FLASHW];
} nvmcfg;

typedef struct {
    uint8_t         channel;                //Analog channel voltage acquired from
    uint8_t         scale;                  //Measurement scale
    uint8_t         mvr;                    //mV/ADC bit: see ADC section for values
    uint16_t        ofs;                    //Measurement offset
    uint16_t        volt;                   //Voltage
    uint16_t        mv;                     //Millivolts (3 decimals)
    uint16_t        vmv;                    //Composite volts & mv represented as integer
    uint16_t        adc;                    //Unmodified value read from ADC
    uint16_t        adc1v;                  //ADC value for 1 volt (match mvr)
} volts;


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * MACROS.
 * + SAFEASSIGN ensures safe assignment of objects in foreground execution that
 *   are larger than the MCU's native data size and are also modified by ISR.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#define SAFEASSIGN(lval, rval)  GIE=FALSE; lval=rval; GIE=TRUE


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Function declarations.  See source file for explanation of each function.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void ReadVoltage(volts *);
void writeFlashRow(uint16_t, unsigned char *);
void eraseFlashRow(uint16_t addr);
uint16_t nvmChksum(unsigned char *);



//End of file
